// -*-c++-*-

/*!
  \file angle_deg.h
  \brief degree wrapper class Header File.
*/

/*
 *Copyright:

 Copyright (C) Hidehisa Akiyama

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef RCSC_GEOM_ANGLEDEG_H
#define RCSC_GEOM_ANGLEDEG_H

#include <iostream>
#include <cmath>
#include <functional>

namespace rcsc {

class AngleDeg {
    // : public boost::addable< AngleDeg >
    // , public boost::addable2< AngleDeg, double >
    // , public boost::addable2_left< double,  AngleDeg >
    // , public boost::subtractable< AngleDeg >
    // , public boost::subtractable2< AngleDeg, double >
    // , public boost::subtractable2_left< double, AngleDeg >
    // , public boost::multipliable2< AngleDeg, double >
    // , public boost::dividable2< AngleDeg, double >
private:
    //! degree, NOT radian
    double M_degree;

public:
    static const double EPSILON;

    //! constant variable to convert DEGREE to RADIAN.
    static const double DEG2RAD;
    //! constant variable to convert RADIAN to DEGREE.
    static const double RAD2DEG;

    //! default constructor.
    AngleDeg()
        : M_degree( 0.0 )
      { }
    /*!
      \brief constructor with value.
      NO explicit.
      This means that we can use this class as follows.
      AngleDeg angle = 3.0;
    */
    AngleDeg( const double & deg )
        : M_degree( deg )
      {
          normalize();
      }

    //! operator substitution.
    inline
    const
    AngleDeg & operator=( const double & deg )
      {
          M_degree = deg;
          return normalize();
      }

private:
    /*!
      \brief normalize the value to [-180, 180].
      \return const reference to this.
    */
    inline
    const
    AngleDeg & normalize()
      {
          while ( M_degree < -180.0 )
          {
              M_degree += 360.0;
          }
          while ( M_degree > 180.0 )
          {
              M_degree -= 360.0;
          }
          return *this;
      }

public:
    //! get const reference to M_degree.
    inline
    const
    double & degree() const
      {
          return M_degree;
      }
    //! get absolute value of M_degree.
    inline
    double abs() const
      {
          return std::fabs( degree() );
      }
    //! get RADIAN value of M_degree.
    inline
    double radian() const
      {
          return degree() * DEG2RAD;
      }

    /*
      operator double() const
      {
      return degree();
      }
    */

    //! get new AngleDeg that M_degree is multiplyed by -1.
    inline
    AngleDeg operator-() const
      {
          return AngleDeg( - degree() );
      }
    //! operator +=
    inline
    const
    AngleDeg & operator+=( const AngleDeg & ang )
      {
          M_degree += ang.degree();
          return normalize();
      }
    //! operator +=
    inline
    const
    AngleDeg & operator+=( const double & deg )
      {
          M_degree += deg;
          return normalize();
      }
    //! operator -= with AngleDeg
    inline
    const
    AngleDeg & operator-=( const AngleDeg & ang )
      {
          M_degree -= ang.degree();
          return normalize();
      }
    //! operator -= with double
    inline
    const
    AngleDeg & operator-=( const double & deg )
      {
          M_degree -= deg;
          return normalize();
      }
    //! operator *=
    inline
    const
    AngleDeg & operator*=( const double & scalar )
      {
          M_degree *= scalar;
          return normalize();
      }
    //! operator /=
    inline
    const
    AngleDeg & operator/=( const double & scalar )
      {
          M_degree /= scalar;
          return normalize();
      }

    //////////////////////////////////////////////////////////////////

    inline
    bool isLeftOf( const AngleDeg & angle ) const
      {
          //return (*this - angle).degree() < 0.0;
          double diff = angle.degree() - this->degree();
          return ( ( 0.0 < diff && diff < 180.0 )
                   || diff < -180.0 );
      }

    inline
    bool isLeftEqualOf( const AngleDeg & angle ) const
      {
          //return (*this - angle).degree() <= 0.0;
          double diff = angle.degree() - this->degree();
          return ( ( 0.0 <= diff && diff < 180.0 )
                   || diff < -180.0 );
      }

    inline
    bool isRightOf( const AngleDeg & angle ) const
      {
          //return (*this - angle).degree() > 0.0;
          double diff = this->degree() - angle.degree();
          return ( ( 0.0 < diff && diff < 180.0 )
                   || diff < -180.0 );
      }

    inline
    bool isRightEqualOf( const AngleDeg & angle ) const
      {
          //return (*this - angle).degree() >= 0.0;
          double diff = this->degree() - angle.degree();
          return ( ( 0.0 <= diff && diff < 180.0 )
                   || diff < -180.0 );
      }

    //! get cosine value
    inline
    double cos() const
      {
          return std::cos( degree() * DEG2RAD );
      }
    //! get sine value
    inline
    double sin() const
      {
          return std::sin( degree() * DEG2RAD );
      }
    //! get tangent value
    inline
    double tan() const
      {
          return std::tan( degree() * DEG2RAD );
      }

    //! check if this is within from left to right <-- turn clockwise
    bool isWithin( const AngleDeg & left, const AngleDeg & right ) const;

    /*!
      \brief get bisect angle, but... *** Use carefully! ***
      \param angle result is the bisect between '*this' to 'angle'
      \return bisect angle
    */
    inline
    AngleDeg bisect( const AngleDeg & angle ) const
      {
          AngleDeg rel( angle );
          rel -= *this;
          return AngleDeg( *this ) += rel.degree() * 0.5;
          //return AngleDeg(this->degree() + ((ang - *this).degree() * 0.5));
      }

    ////////////////////////////////////////////////////////
    // static utility method concerned with angle operation

    /*!
      static utility.
      angle normalization.
      \param dir angle value to normalize
      \return normalized angle value
     */
    inline
    static
    double normalize_angle( double dir )
      {
          while ( dir < -180.0 )
          {
              dir += 360.0;
          }
          while ( dir > 180.0 )
          {
              dir -= 360.0;
          }
          return dir;
      }

    /*!
      \brief static utility.
      convert radian to degree
      \param radx radian value
      \return degree value
    */
    inline
    static
    double rad2deg( const double & rad )
      {
          return rad * RAD2DEG;
      }
    /*!
      \brief static utility.
      convert degree to radian
      \param degx degree value
      \return radian value
    */
    inline
    static
    double deg2rad( const double & deg )
      {
          return deg * DEG2RAD;
      }

    /*!
      \brief static utility.
      calculate cosine value for degree angle
      \param degx degree value
      \return cosine value
    */
    inline
    static
    double cos_deg( const double & deg )
      {
          return std::cos( deg2rad( deg ) );
      }

    /*!
      \brief static utility.
      calculate sine value for degree angle
      \param degx degree value
      \return sine value
    */
    inline
    static
    double sin_deg( const double & deg )
      {
          return std::sin( deg2rad( deg ) );
      }

    /*!
      \brief static utility.
      calculate tangent value for degree angle
      \param degx degree value
      \return tangent value
    */
    inline
    static
    double tan_deg( const double & deg )
      {
          return std::tan( deg2rad( deg ) );
      }

    /*!
      \breif static utility.
      calculate arc cosine value
      \param x cosine value
      \return arc cosine value, but this is degree.
    */
    inline
    static
    double acos_deg( const double & cosine )
      {
          return ( cosine >= 1.0
                   ? 0.0
                   : ( cosine <= -1.0
                       ? 180.0
                       : rad2deg( std::acos( cosine ) ) ) );
      }

    /*!
      \brief static utility.
      calculate arc sine value
      \param x sine value
      \return arc sine value, but this is degree.
    */
    inline
    static
    double asin_deg( const double & sine )
      {
          return ( sine >= 1.0
                   ? 90.0
                   : ( sine <= -1.0
                       ? -90.0
                       : rad2deg( std::asin( sine ) ) ) );
      }

    /*!
      \brief static utility.
      calculate arc tangent value
      \param x tangent value
      \return arc tangent value, but this is degree.
    */
    inline
    static
    double atan_deg( const double & tangent )
      {
          return rad2deg( std::atan( tangent ) );
      }

    /*!
      \brief static utility.
      calculate arc tangent value from XY
      \param y coordinate Y
      \param x coordinate X
      \return arc tangent value, but this is degree.
    */
    inline
    static
    double atan2_deg( const double & y, const double & x )
      {
          return ( ( x == 0.0 && y == 0.0 )
                   ? 0.0
                   : rad2deg( std::atan2( y, x ) ) );
      }

    //! get bisect angle.  [left, right] can take a obtuse angle.
    static
    AngleDeg bisect( const AngleDeg & left, const AngleDeg & right );

    ////////////////////////////////////////////////////////
    //! output value
    inline
    std::ostream & print( std::ostream & o ) const
      {
          return o << degree();
      }
    //! output rounded value
    inline
    std::ostream & printRound( std::ostream & o, const double & step = 0.1 ) const
      {
          return o << rint( degree() / step ) * step;
      }

    ////////////////////////////////////////////////////////
    /*!
      predicate function object.
      this compares two angles by degree value
    */
    class DegreeCmp
        : public std::binary_function< AngleDeg, AngleDeg, bool > {
    public:
        //! operator method
        result_type operator()( const first_argument_type & lhs,
                                const second_argument_type & rhs ) const
          {
              return lhs.degree() < rhs.degree();
          }
    };

};


////////////////////////////////////////////////////////
/*!
  predicate function object.
  this checks if input angle is within specific range.
*/
class AngleIsWithin
    : public std::unary_function< AngleDeg, bool > {
    const AngleDeg M_left; //!< left threshold
    const AngleDeg M_right; //!< right threshold
public:
    //! constructor
    AngleIsWithin( const AngleDeg & left, const AngleDeg & right )
        : M_left( left ), M_right( right )
      { }

    //! operator method
    result_type operator()( const argument_type & angle ) const
      {
          return angle.isWithin( M_left, M_right );
      }
};


} // end of namespace


////////////////////////////////////////////////////////
// arith operators

/*-------------------------------------------------------------------*/
/*!
  \brief operator add(T, T)
  \param lhs left hand side argument
  \param rhs right hand side argument
  \return new angle
 */
inline
const
rcsc::AngleDeg
operator+( const rcsc::AngleDeg & lhs, const rcsc::AngleDeg & rhs )
{
    return rcsc::AngleDeg( lhs ) += rhs;
}

/*-------------------------------------------------------------------*/
/*!
  \brief operator add(T, U)
  \param lhs left hand side argument
  \param rhs right hand side argument
  \return new angle
 */
inline
const
rcsc::AngleDeg
operator+( const rcsc::AngleDeg & lhs, const double & rhs )
{
    return rcsc::AngleDeg( lhs ) += rhs;
}

/*-------------------------------------------------------------------*/
/*!
  \brief operator add(U, T)
  \param lhs left hand side argument
  \param rhs right hand side argument
  \return new angle
 */
inline
const
rcsc::AngleDeg
operator+( const double & lhs, const rcsc::AngleDeg & rhs )
{
    return rcsc::AngleDeg( rhs ) += lhs;
}

/*-------------------------------------------------------------------*/
/*!
  \brief operator sub(T, T)
  \param lhs left hand side argument
  \param rhs right hand side argument
  \return new angle
 */
inline
const
rcsc::AngleDeg
operator-( const rcsc::AngleDeg & lhs, const rcsc::AngleDeg & rhs )
{
    return rcsc::AngleDeg( lhs ) -= rhs;
}

/*-------------------------------------------------------------------*/
/*!
  \brief operator sub(T, U)
  \param lhs left hand side argument
  \param rhs right hand side argument
  \return new angle
 */
inline
const
rcsc::AngleDeg
operator-( const rcsc::AngleDeg & lhs, const double & rhs )
{
    return rcsc::AngleDeg( lhs ) -= rhs;
}

/*-------------------------------------------------------------------*/
/*!
  \brief operator sub(U, T)
  \param lhs left hand side argument
  \param rhs right hand side argument
  \return new angle
 */
inline
const
rcsc::AngleDeg
operator-( const double & lhs, const rcsc::AngleDeg & rhs )
{
    return rcsc::AngleDeg( lhs ) -= rhs;
}

/////////////////////////////////////////////////////
// predicate operator

/*-------------------------------------------------------------------*/
/*!
  \brief predicate operator !=
  \param lhs left hand side argument
  \param rhs right hand side argument
  \return true or false
 */
inline
bool
operator!=( const rcsc::AngleDeg & lhs, const rcsc::AngleDeg & rhs )
{
    return std::fabs( lhs.degree() - rhs.degree() ) > rcsc::AngleDeg::EPSILON;
}

/*-------------------------------------------------------------------*/
/*!
  \brief predicate operator !=
  \param lhs left hand side argument
  \param rhs right hand side argument
  \return true or false
 */
inline
bool
operator!=( const rcsc::AngleDeg & lhs, const double & rhs )
{
    return std::fabs( lhs.degree() - rhs ) > rcsc::AngleDeg::EPSILON;
}

/*-------------------------------------------------------------------*/
/*!
  \brief predicate operator !=
  \param lhs left hand side argument
  \param rhs right hand side argument
  \return true or false
 */
inline
bool
operator!=( const double & lhs, const rcsc::AngleDeg & rhs )
{
    return std::fabs( lhs - rhs.degree() ) > rcsc::AngleDeg::EPSILON;
}

/*-------------------------------------------------------------------*/
/*!
  \brief predicate operator ==
  \param lhs left hand side argument
  \param rhs right hand side argument
  \return true or false
 */
inline
bool
operator==( const rcsc::AngleDeg & lhs, const rcsc::AngleDeg & rhs )
{
    return std::fabs( lhs.degree() - rhs.degree() ) < rcsc::AngleDeg::EPSILON;
}

/*-------------------------------------------------------------------*/
/*!
  \brief predicate operator ==
  \param lhs left hand side argument
  \param rhs right hand side argument
  \return true or false
 */
inline
bool
operator==( const rcsc::AngleDeg& lhs, const double& rhs )
{
    return std::fabs( lhs.degree() - rhs ) < rcsc::AngleDeg::EPSILON;
}

/*-------------------------------------------------------------------*/
/*!
  \brief predicate operator ==
  \param lhs left hand side argument
  \param rhs right hand side argument
  \return true or false
 */
inline
bool
operator==( const double & lhs, const rcsc::AngleDeg & rhs )
{
    return std::fabs( lhs - rhs.degree() ) < rcsc::AngleDeg::EPSILON;
}

/////////////////////////////////////////////////////
// stream operator

/*-------------------------------------------------------------------*/
/*!
  output to stream
  \param o reference to the output stream
  \param a angle object
  \return reference to the output stream
 */
inline
std::ostream &
operator<<( std::ostream & o, const rcsc::AngleDeg & a )
{
    return a.print( o );
}

#endif
